import unittest

from functional.chain import seq, FunctionalSequence


class TestChain(unittest.TestCase):
    def assertType(self, s):
        self.assertTrue(isinstance(s, FunctionalSequence))

    def test_eq(self):
        l = [1, 2, 3]
        self.assertEqual(seq(l), seq(l))

    def test_ne(self):
        a = [1, 2, 3]
        b = [1]
        self.assertNotEqual(seq(a), seq(b))

    def test_repr(self):
        l = [1, 2, 3]
        self.assertEqual(repr(l), repr(seq(l)))

    def test_str(self):
        l = [1, 2, 3]
        self.assertEqual(str(l), str(seq(l)))

    def test_len(self):
        l = [1, 2, 3]
        s = seq(l)
        self.assertEqual(len(l), len(s))
        self.assertEqual(len(l), s.count())
        self.assertEqual(len(l), s.size())
        self.assertEqual(len(l), s.len())

    def test_map(self):
        f = lambda x: x * 2
        l = [1, 2, 0, 5]
        expect = [2, 4, 0, 10]
        result = seq(l).map(f)
        self.assertEqual(expect, result)
        self.assertType(result)

    def test_filter(self):
        f = lambda x: x > 0
        l = [0, -1, 5, 10]
        expect = [5, 10]
        s = seq(l)
        result = s.filter(f)
        self.assertEqual(expect, result)
        self.assertType(result)

    def test_filter_not(self):
        f = lambda x: x > 0
        l = [0, -1, 5, 10]
        expect = [0, -1]
        result = seq(l).filter_not(f)
        self.assertEqual(expect, result)
        self.assertType(result)

    def test_map_filter(self):
        f = lambda x: x > 0
        g = lambda x: x * 2
        l = [0, -1, 5]
        s = seq(l)
        expect = [10]
        result = s.filter(f).map(g)
        self.assertEqual(expect, result)
        self.assertType(result)

    def test_reduce(self):
        f = lambda x, y: x + y
        l = ["a", "b", "c"]
        expect = "abc"
        s = seq(l)
        self.assertEqual(expect, s.reduce(f))

    def test_sum(self):
        l = [1, 2, 3]
        self.assertEqual(6, seq(l).sum())

    def test_reverse(self):
        l = [1, 2, 3]
        expect = [3, 2, 1]
        s = seq(l)
        result = s.reverse()
        self.assertEqual(expect, result)
        self.assertType(result)
        result = s.__reversed__()
        self.assertEqual(expect, result)
        self.assertType(result)

    def test_distinct(self):
        l = [1, 1, 2, 3, 2, 3]
        expect = [1, 2, 3]
        s = seq(l)
        result = s.distinct()
        for e in result:
            self.assertTrue(e in expect)
        result = s.distinct()
        self.assertEqual(result.size(), len(expect))
        self.assertType(result)

    def test_any(self):
        l = [True, False]
        self.assertTrue(seq(l).any())

    def test_all(self):
        l = [True, False]
        self.assertFalse(seq(l).all())
        l = [True, True]
        self.assertTrue(seq(l).all())

    def test_max(self):
        l = [1, 2, 3]
        self.assertEqual(3, seq(l).max())

    def test_min(self):
        l = [1, 2, 3]
        self.assertEqual(1, seq(l).min())

    def test_find(self):
        l = [1, 2, 3]
        f = lambda x: x == 3
        g = lambda x: False
        self.assertEqual(3, seq(l).find(f))
        self.assertIsNone(seq(l).find(g))

    def test_flat_map(self):
        l = [[1, 1, 1], [2, 2, 2], [3, 3, 3]]
        f = lambda x: x
        expect = [1, 1, 1, 2, 2, 2, 3, 3, 3]
        result = seq(l).flat_map(f)
        self.assertEqual(expect, result)
        self.assertType(result)

    def test_group_by(self):
        l = [(1, 1), (1, 2), (1, 3), (2, 2)]
        f = lambda x: x[0]
        expect = {1: [(1, 1), (1, 2), (1, 3)], 2: [(2, 2)]}
        result = seq(l).group_by(f)
        self.assertEqual(expect, result)
        self.assertType(result)

    def test_empty(self):
        self.assertTrue(seq([]).empty())

    def test_non_empty(self):
        self.assertTrue(seq([1]).non_empty())

    def test_string(self):
        l = [1, 2, 3]
        expect1 = "1 2 3"
        expect2 = "1:2:3"
        s = seq(l)
        self.assertEqual(expect1, s.string(" "))
        self.assertEqual(expect2, s.string(":"))

    def test_partition(self):
        l = [-1, -2, -3, 1, 2, 3]
        e2 = [-1, -2, -3]
        e1 = [1, 2, 3]
        f = lambda x: x > 0
        s = seq(l)
        p1, p2 = s.partition(f)
        self.assertEqual(e1, p1)
        self.assertEqual(e2, p2)
        self.assertType(p1)
        self.assertType(p2)

    def test_product(self):
        l = [2, 2, 3]
        self.assertEqual(12, seq(l).product())

    def test_set(self):
        l = [1, 1, 2, 2, 3]
        ls = set(l)
        self.assertEqual(ls, seq(l).set())

    def test_zip(self):
        l1 = [1, 2, 3]
        l2 = [-1, -2, -3]
        e = [(1, -1), (2, -2), (3, -3)]
        result = seq(l1).zip(l2)
        self.assertEqual(e, result)
        self.assertType(result)

    def test_zip_with_index(self):
        l = [2, 3, 4]
        e = [(0, 2), (1, 3), (2, 4)]
        result = seq(l).zip_with_index()
        self.assertEqual(result, e)
        self.assertType(result)

    def test_enumerate(self):
        l = [2, 3, 4]
        e = [(0, 2), (1, 3), (2, 4)]
        result = seq(l).enumerate()
        self.assertEqual(result, e)
        self.assertType(result)



class FunctionalSequence:
    def __init__(self, sequence):
        if isinstance(sequence, list) or isinstance(sequence, dict):
            self.sequence = sequence
        else:
            self.sequence = list(sequence)

    def __eq__(self, other):
        return self.sequence == other

    def __ne__(self, other):
        return self.sequence != other

    def __hash__(self):
        return hash(self.sequence)

    def __repr__(self):
        return repr(self.sequence)

    def __str__(self):
        return str(self.sequence)

    def __unicode__(self):
        return unicode(self.sequence)

    def __format__(self, formatstr):
        return format(self.sequence, formatstr)

    def __nonzero__(self):
        return self.size() != 0

    def __len__(self):
        return len(self.sequence)

    def __getitem__(self, key):
        return self.sequence[key]

    def __setitem__(self, key, value):
        self.sequence[key] = value

    def __delitem__(self, key):
        del self.sequence[key]

    def __iter__self(self):
        return iter(self.sequence)

    def __reversed__(self):
        return FunctionalSequence(reversed(self.sequence))

    def __contains__(self, item):
        return self.sequence.__contains__(item)

    def head(self):
        return FunctionalSequence(self.sequence[0])

    def first(self):
        return FunctionalSequence(self.sequence[0])

    def last(self):
        return FunctionalSequence(self.sequence[-1])

    def tail(self):
        return FunctionalSequence(self.sequence[-1])

    def drop(self, n):
        return FunctionalSequence(self.sequence[n:])

    def take(self, n):
        return FunctionalSequence(self.sequence[:n])

    def map(self, f):
        return FunctionalSequence(map(f, self.sequence))

    def filter(self, f):
        return FunctionalSequence(filter(f, self.sequence))

    def filter_not(self, f):
        g = lambda x: not f(x)
        return FunctionalSequence(filter(g, self.sequence))

    def reduce(self, f):
        return reduce(f, self.sequence)

    def count(self):
        return len(self.sequence)

    def len(self):
        return len(self.sequence)

    def size(self):
        return len(self.sequence)

    def reverse(self):
        return reversed(self)

    def distinct(self):
        return FunctionalSequence(list(set(self.sequence)))

    def any(self):
        return any(self.sequence)

    def all(self):
        return all(self.sequence)

    def enumerate(self, start=0):
        return FunctionalSequence(enumerate(self.sequence, start=start))

    def max(self):
        return max(self.sequence)

    def min(self):
        return min(self.sequence)

    def find(self, f):
        for e in self.sequence:
            if f(e):
                return e
        else:
            return None

    def flat_map(self, f):
        l = []
        for e in self.sequence:
            l.extend(f(e))
        return FunctionalSequence(l)

    def group_by(self, f):
        result = {}
        for e in self.sequence:
            if result.get(f(e)):
                result.get(f(e)).append(e)
            else:
                result[f(e)] = [e]
        return FunctionalSequence(result)

    def empty(self):
        return len(self.sequence) == 0

    def non_empty(self):
        return len(self.sequence) != 0

    def string(self, separator):
        if self.empty():
            return ""
        first = True
        s = ""
        for e in self.sequence:
            if first:
                s += str(e)
                first = False
            else:
                s += separator + str(e)
        return s

    def partition(self, f):
        p1 = self.filter(f)
        p2 = self.filter_not(f)
        return p1, p2

    def product(self):
        return self.reduce(lambda x, y: x * y)

    def slice(self, start, until):
        return FunctionalSequence(self.sequence[start:until])

    def sum(self):
        return sum(self.sequence)

    def set(self):
        return set(self.sequence)

    def zip(self, sequence):
        return FunctionalSequence(zip(self.sequence, sequence))

    def zip_with_index(self):
        return FunctionalSequence(list(enumerate(self.sequence)))

    def sorted(self, comp=None, key=None, reverse=False):
        return FunctionalSequence(sorted(self.sequence, cmp=comp, key=key, reverse=reverse))

    def list(self):
        return self.sequence


def seq(l):
    return FunctionalSequence(l)


from setuptools import setup, find_packages

setup(
    name='ScalaFunctional',
    description="Scala functional programming style in Python",
    url="https://github.com/EntilZha/ScalaFunctional",
    author="Pedro Rodriguez",
    author_email="pedro@snowgeek.org",
    license="MIT",
    keywords="functional scala",
    packages=find_packages(exclude=['contrib', 'docs', 'tests*', 'test']),
    version="0.0.1",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7"
    ]
)